-- == Custom HUD Functions by TehRealSalt ==
-- Global script that contains extra functions for any HUD-related scripts.

-- USAGE: Main part of customhud is the ability to creating and overwriting existing HUD items, and adding support for other HUD modifications.
-- This helps sort out HUD conflicts that are otherwise impossible to detect without the use of this library.

-- customhud.SetupItem(itemname, modname) lets you create a new HUD item. "itemname" is the identifier for a HUD item, while "modname" is the identifier for what it belongs to so it can be overwritten. Please run only once on script load.

-- Examples: customhud.SetupItem("bossmeter", "ktdbossmeter") registers the HUD item "bossmeter" to the mod "ktdbossmeter"
--			 (If bossmeter already exists, then the modname of bossmeter will be changed to ktdbossmeter anyways.)
--			 customhud.SetupItem("rings", "coolrings") overwrites the ring counter that's loaded (typically vanilla's), and registers it to the mod "coolrings".
--			 customhud.SetupItem("lives", "vanilla") changes the lives counter back to normal. ("vanilla" is the modname reserved for HUD items that exist in normal SRB2, and cannot be set to new HUD items.)

-- After a new HUD item is set up, it can be enabled or disabled with customhud.enable(itemname) and customhud.disable(itemname), respectively.
-- You can determine if your HUD item should be rendered by checking if customhud.CheckEnabled(itemname) is true and customhud.CheckType(itemname) is your modname.
-- Other people can also support your HUD replacements by checking for the same things.
-- Beyond that, customhud.ItemExists can also be used to check if a HUD item already exists or not.

-- This also comes with some extra drawing functions.
-- customhud.CustomFontString(v, x, y, str, name, space, width, flags, align, scale, color) draws a string using a different font.
-- "v" is the drawer variable supplied by a HUD hook. "x" and "y" are the coordinates to draw at. "str" is the string to draw.
-- "name" is the font name prefix to use (must be between 1-5 characters long). Each lump should be stored in a WAD as "YYYYYxxx", with the Y's being the font name and the x's being the byte of the character.
-- "space" is the spacing to use when a character doesn't exist. Defaults to 4 if omitted.
-- "width" is the spacing between characters that do exist. Set to nil to use each individual patch's width instead.
-- "flags" are the flags used for each patch. "align" is how it should be aligned (can be "left", "center", or "right").
-- "scale" rescales the string if set. "color" is the SKINCOLOR_ constant to remap the text to.

-- Examples: customhud.CustomFontString(v, 8, 16, "Hello World!", "CRFNT") draws a string using SRB2's credits font
--			 customhud.CustomFontString(v, 16*FRACUNIT, 8*FRACUNIT, "World, Hello!", "LTFNT", nil, nil, 0, "left", 2*FRACUNIT) draws a rescaled string using SRB2's level title font
--			 customhud.CustomFontString(v, 16, 8, "Ayy, The World!", "COOLF", 8, 8, 0, "left", nil, SKINCOLOR_BLUE) draws a recolored, monospaced string using a custom font

-- NOTE: customhud.CustomFontString technically just draws patches for you. This means string-only flags will not work, patch-only flags will work, and green instead of white is remapped to the color.

-- customhud.CustomFontStringWidth(v, str, name, space, width, scale) returns the width of a custom font string. Each variable is identical to customhud.CustomFontString's.

-- customhud.CustomNum(v, x, y, str, name, space, width, padding, flags, align, scale, color) is similar to customhud.CustomFontString, except draws a font made of only numbers.
-- The font's lumps should be stored in a WAD as "YYYYYYYx", with the Y's being the font name and the x being the number (0-9).
-- padding is the number of zeroes to pad with.

-- customhud.CustomNumWidth(v, str, name, space, width, padding, scale) returns the width of a custom font number. Each variable is identical to customhud.CustomNum's.

-- Lastly, customhud.GetVersionNum() grabs the version number of this library, in table format. Usually, if you need to check this, you'll only need to check the first entry.

-- DO NOT EDIT THIS FILE! Don't add your custom HUD elements in here! Do it with the customhud.SetupItem function in another script instead.
-- If you want to add a feature to this, please share your changes with TehRealSalt so it can be put in the main file.

local VERSIONNUM = {1, 3};

local function warn(str)
	print("\131WARNING: \128"..str);
end

local function notice(str)
	print("\x83NOTICE: \x80"..str);
end

if (rawget(_G, "customhud")) then
	local oldnum = customhud.GetVersionNum();
	local numlength = max(#VERSIONNUM, #oldnum);

	local loadednum = "";
	local newnum = "";

	local oldvers = false;

	for i = 1,numlength
		local num1 = oldnum[i];
		local num2 = VERSIONNUM[i];

		if (num1 == nil) then
			num1 = 0;
		end
		if (num2 == nil) then
			num2 = 0;
		end

		if (loadednum == "") then
			loadednum = "v"..num1;
		else
			loadednum = $1.."."..num1;
		end

		if (newnum == "") then
			newnum = "v"..num2;
		else
			newnum = $1.."."..num2;
		end

		if (num1 < num2) then
			oldvers = true;
		elseif (num1 > num2) then
			break;
		end
	end

	if (oldvers == false) then
		-- Existing version is OK
		return;
	end

	notice("An old version of customhud was detected ("..loadednum.."). Switching to newer ("..newnum.."), errors may occur.");
end

rawset(_G, "customhud", {});

function customhud.GetVersionNum()
	-- Make sure you cannot overwrite the version number by copying it into another table
	-- That'd be really silly :V
	local tempNum = {};

	for k,v in ipairs(VERSIONNUM)
		tempNum[k] = v;
	end

	return tempNum;
end

local huditems = {};

-- Should match hud_disable_options in lua_hudlib.c
local defaultitems = {
	"stagetitle",
	"textspectator",
	--"crosshair",

	"score",
	"time",
	"rings",
	"lives",

	"weaponrings",
	"powerstones",
	"teamscores",

	"nightslink",
	"nightsdrill",
	"nightsrings",
	"nightsscore",
	"nightstime",
	"nightsrecords",

	"rankings",
	"coopemeralds",
	"tokens",
	"tabemblems",

	"intermissiontally",
	"intermissionmessages",
};

for _,v in pairs(defaultitems)
	huditems[v] = {
		-- Current mod name for the hud item.
		-- ("vanilla" is a reserved mod name for reverting a modded hud item back to its original state, cannot be set for custom hud items)
		type = "vanilla",
		-- Is this hud item supposed to be visible?
		enabled = hud.enabled(v),
		-- Determines if this is a default HUD item defined by the game. Should never be true for custom HUD elements.
		isDefaultItem = true
	};
end

local function FindItem(item)
	for k,_ in pairs(huditems)
		if (k == item) then
			return huditems[k];
		end
	end

	return nil;
end

function customhud.ItemExists(item)
	if (FindItem(item) != nil) then
		return true;
	else
		return false;
	end
end

function customhud.enable(item)
	local x = FindItem(item);
	if (x == nil) then
		return false;
	end

	x.enabled = true;

	if (x.isDefaultItem == true) then
		if (x.type == "vanilla") then
			hud.enable(item);
		else
			hud.disable(item);
		end
	end

	return true;
end

function customhud.disable(item)
	local x = FindItem(item);
	if (x == nil) then
		return false;
	end

	x.enabled = false;

	if (x.isDefaultItem == true) then
		hud.disable(item);
	end

	return true;
end

function customhud.CheckEnabled(item)
	local x = FindItem(item);
	if (x == nil) then
		return false;
	end

	if (x.isDefaultItem == true)
	and (x.type == "vanilla") then
		return hud.enabled(item);
	end

	return x.enabled;
end

function customhud.CheckType(item)
	local x = FindItem(item);
	if (x == nil) then
		return nil;
	end

	return x.type;
end

function customhud.SetupItem(item, t)
	if not (type(t) == "string") then
		warn("Invalid type string \""..t.."\" in customhud.SetupItem");
		return;
	end

	local x = FindItem(item);

	if (x == nil)
		-- Create new item
		huditems[item] = {
			type = t,
			enabled = true,
			isDefaultItem = false
		};

		return true;
	end

	-- Updating existing item
	if (x.type == t) then
		-- Already set to this type.
		return false;
	end

	if (t == "vanilla") and (x.isDefaultItem != true) then
		-- Trying to set a custom HUD item to "vanilla".
		warn("Type string \"vanilla\" is reserved for specific hud items in customhud.SetupItem")
		return false;
	end

	x.type = t;

	-- Update status
	if (x.enabled == false) then
		hud.disable(item);
	elseif (x.isDefaultItem == true) then
		if (t == "vanilla") then
			hud.enable(item);
		else
			hud.disable(item);
		end
	end

	return true;
end

local function pname(name, wb)
	return name..string.format("%03d", wb);
end

function customhud.CustomFontStringWidth(v, str, name, space, width, scale)
	if not (type(str) == "string") then
		warn("No string given in customhud.CustomFontStringWidth");
		return;
	end

	if not (type(name) == "string") then
		warn("No font given in customhud.CustomFontStringWidth");
		return;
	end

	if (name:find(" ")) then
		warn("Font name \""..name.."\" cannot have spaces in customhud.CustomFontStringWidth");
		return;
	end

	if (name:len() > 5) or (name:len() < 1) then
		warn("Incorrect font name length in customhud.CustomFontStringWidth");
		return;
	end

	local strwidth = 0;

	if (str == "") then
		return strwidth;
	end

	if not (type(space) == "number") then
		space = 4;

		if (type(scale) == "number") then
			space = $ * scale;
		end
	end

	local nextx = x;
	local capsoffset = 32;

	for i = 1,str:len() do
		local nextbyte = str:byte(i,i);
		local patchn = pname(name,nextbyte);

		if not (v.patchExists(patchn)) then
			if (nextbyte >= 65 and nextbyte <= 90)
				nextbyte = $ + capsoffset;
				patchn = pname(name,nextbyte);
			elseif (nextbyte >= 97 and nextbyte <= 122)
				nextbyte = $ - capsoffset;
				patchn = pname(name,nextbyte);
			end
		end

		if (v.patchExists(patchn)) then
			if (type(width) == "number") then
				if (width < 0)
					strwidth = $ + v.cachePatch(patchn).width + width;
				else
					strwidth = $ + width;
				end
			else
				strwidth = $ + v.cachePatch(patchn).width;
			end
		else
			strwidth = $ + space;
		end
	end

	if (type(scale) == "number") then
		strwidth = $ * scale;
	end

	return strwidth;
end

function customhud.CustomFontChar(v, x, y, charByte, name, space, width, flags, scale, color)
	if not (type(charByte) == "number") then
		warn("No character byte given in customhud.CustomFontChar");
		return;
	end

	if not (type(name) == "string") then
		warn("No font given in customhud.CustomFontChar");
		return;
	end

	if (name:find(" ")) then
		warn("Font name \""..name.."\" cannot have spaces in customhud.CustomFontChar");
		return;
	end

	if (name:len() > 5) or (name:len() < 1) then
		warn("Bad font name length in customhud.CustomFontChar");
		return;
	end

	if not (type(space) == "number") then
		space = 4;

		if (type(scale) == "number") then
			space = $ * scale;
		end
	end

	local wc = nil;
	if (color) then
		wc = v.getColormap(TC_DEFAULT, color);
	end

	local patchn = pname(name,charByte);
	local patch = nil;

	local capsoffset = 32;
	if not (v.patchExists(patchn))
		if ((charByte >= 65 and charByte <= 90)
		and (v.patchExists(pname(name, charByte + 32)))) then
			charByte = $ + capsoffset;
			patchn = pname(name,charByte);
		elseif ((charByte >= 97 and charByte <= 122)
		and (v.patchExists(pname(name, charByte - 32)))) then
			charByte = $ - capsoffset;
			patchn = pname(name,charByte);
		end
	end

	if (v.patchExists(patchn)) then
		patch = v.cachePatch(patchn);

		if (type(scale) == "number") then
			v.drawScaled(x, y, scale, patch, flags, wc);
		else
			v.draw(x, y, patch, flags, wc);
		end
	end

	local nextx = x;
	if (patch and patch.valid) then -- If the character exists, move the next patch over to the right by...
		if (type(width) == "number") then
			if (width < 0)
				if (type(scale) == "number") then
					nextx = $ + ((patch.width + width) * scale);
				else
					nextx = $ + (patch.width + width);
				end
			else
				if (type(scale) == "number") then
					nextx = $ + (width * scale);
				else
					nextx = $ + width;
				end
			end
		else
			if (type(scale) == "number") then
				nextx = $ + (patch.width * scale);
			else
				nextx = $ + patch.width;
			end
		end
	else
		nextx = $ + space;
	end

	return nextx;
end

function customhud.CustomFontString(v, x, y, str, name, space, width, flags, align, scale, color)
	if not (type(str) == "string") then
		warn("No string given in customhud.CustomFontString");
		return;
	end

	if not (type(name) == "string") then
		warn("No font given in customhud.CustomFontString");
		return;
	end

	if (name:find(" ")) then
		warn("Font name \""..name.."\" cannot have spaces in customhud.CustomFontString");
		return;
	end

	if (name:len() > 5) or (name:len() < 1) then
		warn("Bad font name length in customhud.CustomFontString");
		return;
	end

	if (str == "") then
		return;
	end

	local nextx = x;

	if (align == "right" or align == "center") then
		local strwidth = customhud.CustomFontStringWidth(v, str, name, space, width, scale);

		if (align == "right") then
			nextx = x - strwidth;
		elseif (align == "center") then
			nextx = x - (strwidth / 2);
		end
	end

	for i = 1,str:len() do
		local nextbyte = str:byte(i,i);
		nextx = customhud.CustomFontChar(v, nextx, y, nextbyte, name, space, width, flags, scale, color);
	end
end

function customhud.CustomNumWidth(v, num, name, width, pad, scale)
	if not (type(num) == "number") then
		warn("No number given in customhud.CustomNumWidth");
		return;
	end

	if not (type(name) == "string") then
		warn("No font given in customhud.CustomNumWidth");
		return;
	end

	if (name:find(" ")) then
		warn("Font name \""..name.."\" cannot have spaces in customhud.CustomNumWidth");
		return;
	end

	if (name:len() > 7) or (name:len() < 1) then
		warn("Incorrect font name length in customhud.CustomNumWidth");
		return;
	end

	local numwidth = 0;

	num = tostring($1);
	if (tonumber(pad)) then
		num = string.format("%0"..pad.."d", $1);
	end

	if (num == "") then
		return numwidth;
	end

	for i = 1,num:len() do
		local nextnum = num:sub(i,i);

		if (v.patchExists(name..nextnum)) then
			if (type(width) == "number") then
				numwidth = $ + width;
			else
				numwidth = $ + v.cachePatch(name..nextnum).width;
			end
		end
	end

	if (type(scale) == "number") then
		numwidth = $ * scale;
	end

	return numwidth;
end

function customhud.CustomNum(v, x, y, num, name, width, pad, flags, align, scale, color)
	if not (type(num) == "number") then
		warn("No number given in customhud.CustomNum");
		return;
	end

	if not (type(name) == "string") then
		warn("No font given in customhud.CustomNum");
		return;
	end

	if (name:find(" ")) then
		warn("Font name \""..name.."\" cannot have spaces in customhud.CustomNum");
		return;
	end

	if (name:len() > 7) or (name:len() < 1) then
		warn("Incorrect font name length in customhud.CustomNum");
		return;
	end

	local wc = nil;
	if (color) then
		wc = v.getColormap(TC_DEFAULT, color);
	end

	local nextx = x;

	if (align == "right" or align == "center") then
		local numwidth = customhud.CustomNumWidth(v, num, name, width, pad, scale);

		if (align == "right") then
			nextx = x - numwidth;
		elseif (align == "center") then
			nextx = x - (numwidth / 2);
		end
	end

	num = tostring($1);
	if (tonumber(pad)) then
		num = string.format("%0"..pad.."d", $1);
	end

	for i = 1,num:len() do
		local nextnum = num:sub(i,i);
		local patch = nil;

		if (v.patchExists(name..nextnum)) then
			patch = v.cachePatch(name..nextnum);

			if (type(scale) == "number") then
				v.drawScaled(nextx, y, scale, patch, flags, wc);
			else
				v.draw(nextx, y, patch, flags, wc);
			end
		end

		if (patch and patch.valid) then -- If the character exists, move the next patch over to the right by...
			if (type(width) == "number") then
				if (width < 0)
					if (type(scale) == "number") then
						nextx = $ + ((patch.width + width) * scale);
					else
						nextx = $ + (patch.width + width);
					end
				else
					if (type(scale) == "number") then
						nextx = $ + (width * scale);
					else
						nextx = $ + width;
					end
				end
			else
				if (type(scale) == "number") then
					nextx = $ + (patch.width * scale);
				else
					nextx = $ + patch.width;
				end
			end
		else
			nextx = $ + width;
		end
	end
end
